using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Reflection;

namespace ClassLibrary1
{
    public interface IEntity { }

    public interface ICustomer : IEntity { }

    public interface IValidator<T> where T : IEntity {}

    public class CustomerValidator : IValidator<ICustomer> { }

    public class GenericValidator : IValidator<IEntity> { }


    [TestClass]
    public class Test
    {
        [TestMethod]
        public void ObviouslyShouldNotBeSupportedBy()
        {
            Assert.IsFalse(this.DoesTypeSupportSpecification(typeof(IValidator<IEntity>), typeof(IEntity)));
            Assert.IsFalse(this.DoesTypeSupportSpecification(typeof(IValidator<IEntity>), typeof(ICustomer)));
            Assert.IsFalse(this.DoesTypeSupportSpecification(typeof(IEntity), typeof(IValidator<IEntity>)));
            Assert.IsFalse(this.DoesTypeSupportSpecification(typeof(ICustomer), typeof(CustomerValidator)));
        }

        [TestMethod]
        public void IValidatorOfICustomerShouldNotBeSupportedByGenericValidator()
        {
            Assert.IsFalse(this.DoesTypeSupportSpecification(typeof(IValidator<ICustomer>), typeof(GenericValidator)));
        }

        [TestMethod]
        public void IEntityShouldBeSupportedByIEntity()
        {
            Assert.IsTrue(this.DoesTypeSupportSpecification(typeof(IEntity), typeof(IEntity)));
        }

        [TestMethod]
        public void IEntityShouldBeSupportedByICustomer()
        {
            Assert.IsTrue(this.DoesTypeSupportSpecification(typeof(IEntity), typeof(ICustomer)));
        }

        [TestMethod]
        public void ICustomerShouldNotBeSupportedByIEntity()
        {
            Assert.IsFalse(this.DoesTypeSupportSpecification(typeof(ICustomer), typeof(IEntity)));
        }

        [TestMethod]
        public void IValidatorOfIEntityShouldBeSupportedByIValidatorOfICustomer()
        {
            Assert.IsTrue(this.DoesTypeSupportSpecification(typeof(IValidator<IEntity>), typeof(IValidator<ICustomer>)));
        }

        [TestMethod]
        public void IValidatorOfIEntityShouldBeSupportedByCustomerValidator()
        {
            Assert.IsTrue(this.DoesTypeSupportSpecification(typeof(IValidator<IEntity>), typeof(CustomerValidator)));
        }

        [TestMethod]
        public void IValidatorOfICustomerShouldBeSupportedByCustomerValidator()
        {
            Assert.IsTrue(this.DoesTypeSupportSpecification(typeof(IValidator<ICustomer>), typeof(CustomerValidator)));
        }

        [TestMethod]
        public void IValidatorOfIEntityShouldBeSupportedByGenericValidator()
        {
            Assert.IsTrue(this.DoesTypeSupportSpecification(typeof(IValidator<IEntity>), typeof(GenericValidator)));
        }

        public bool DoesTypeSupportSpecification(Type spec, Type toCheck)
        {
            if (!spec.IsGenericType)
                return spec.IsAssignableFrom(toCheck);

            if (!toCheck.IsGenericType)
            {
                foreach (Type i in toCheck.GetInterfaces())
                    if (DoesTypeSupportSpecification(spec, i))
                        return true;

                return false;
            }

            Type a = spec.GetGenericTypeDefinition();
            Type b = toCheck.GetGenericTypeDefinition();

            if (a != b)
                return false;

            Type[] specArgs = spec.GetGenericArguments();
            Type[] toCheckArgs = toCheck.GetGenericArguments();

            if (specArgs.Length != toCheckArgs.Length)
                return false;

            for (int i = 0; i < specArgs.Length; i++)
                if (!DoesTypeSupportSpecification(specArgs[i], toCheckArgs[i]))
                    return false;
                
            return true;
        }
    }

}
